{ const_cast<char *>("-a"), glkunix_arg_NumberValue, const_cast<char *>("Set the evaluation stack size") },
{ const_cast<char *>("-A"), glkunix_arg_NumberValue, const_cast<char *>("Set the call stack size") },
{ const_cast<char *>("-c"), glkunix_arg_NoValue, const_cast<char *>("Disable colors") },
{ const_cast<char *>("-C"), glkunix_arg_NoValue, const_cast<char *>("Ignore the configuration file") },
{ const_cast<char *>("-d"), glkunix_arg_NoValue, const_cast<char *>("Disable timed input") },
{ const_cast<char *>("-D"), glkunix_arg_NoValue, const_cast<char *>("Disable sound effects") },
{ const_cast<char *>("-e"), glkunix_arg_NoValue, const_cast<char *>("Enable ANSI escape sequences in the transcript") },
{ const_cast<char *>("-E"), glkunix_arg_NoValue, const_cast<char *>("Set the escape string used to decorate transcripts (via -e)") },
{ const_cast<char *>("-f"), glkunix_arg_NoValue, const_cast<char *>("Disable fixed-width fonts") },
{ const_cast<char *>("-F"), glkunix_arg_NoValue, const_cast<char *>("Assume the default font is fixed-width") },
{ const_cast<char *>("-g"), glkunix_arg_NoValue, const_cast<char *>("Disable the character graphics font") },
{ const_cast<char *>("-G"), glkunix_arg_NoValue, const_cast<char *>("Select alternate box-drawing characters") },
{ const_cast<char *>("-h"), glkunix_arg_NoValue, const_cast<char *>("Display this help message then exit") },
{ const_cast<char *>("-H"), glkunix_arg_NoValue, const_cast<char *>("Disable history playback on restore") },
{ const_cast<char *>("-i"), glkunix_arg_NoValue, const_cast<char *>("Print the ID of the specified game then exit") },
{ const_cast<char *>("-k"), glkunix_arg_NoValue, const_cast<char *>("Disable terminating keys") },
{ const_cast<char *>("-l"), glkunix_arg_ValueFollows, const_cast<char *>("Set the username") },
{ const_cast<char *>("-m"), glkunix_arg_NoValue, const_cast<char *>("Disable meta commands") },
{ const_cast<char *>("-n"), glkunix_arg_NumberValue, const_cast<char *>("Set the interpreter number (see 11.1.3 in The Z-machine Standards Document 1.1)") },
{ const_cast<char *>("-N"), glkunix_arg_ValueFollows, const_cast<char *>("Set the interpreter version to the single-character version (see 11.1.3.1 in The Z-machine Standards Document 1.1)") },
{ const_cast<char *>("-p"), glkunix_arg_NoValue, const_cast<char *>("Disable the application of patches") },
{ const_cast<char *>("-r"), glkunix_arg_NoValue, const_cast<char *>("Play back a command record") },
{ const_cast<char *>("-R"), glkunix_arg_ValueFollows, const_cast<char *>("Specify the filename for the command record replay") },
{ const_cast<char *>("-s"), glkunix_arg_NoValue, const_cast<char *>("Turn on command recording") },
{ const_cast<char *>("-S"), glkunix_arg_ValueFollows, const_cast<char *>("Specify the filename for the command record") },
{ const_cast<char *>("-t"), glkunix_arg_NoValue, const_cast<char *>("Turn on transcripting") },
{ const_cast<char *>("-T"), glkunix_arg_ValueFollows, const_cast<char *>("Specify the filename for the transcript") },
{ const_cast<char *>("-u"), glkunix_arg_NumberValue, const_cast<char *>("Set the maximum number of UNDO slots") },
{ const_cast<char *>("-v"), glkunix_arg_NoValue, const_cast<char *>("Display version and configuration information then exit") },
{ const_cast<char *>("-x"), glkunix_arg_NoValue, const_cast<char *>("Disable abbreviations (x, g, z, o) provided by Bocfel for old Infocom games") },
{ const_cast<char *>("-X"), glkunix_arg_NoValue, const_cast<char *>("Turn on the Tandy/censorship flag") },
{ const_cast<char *>("-y"), glkunix_arg_NoValue, const_cast<char *>("Overwrite transcripts rather than appending to them") },
{ const_cast<char *>("-Y"), glkunix_arg_NoValue, const_cast<char *>("Use internal UNDO handling code only") },
{ const_cast<char *>("-z"), glkunix_arg_NumberValue, const_cast<char *>("Provide a seed for the PRNG") },
{ const_cast<char *>("-Z"), glkunix_arg_ValueFollows, const_cast<char *>("Provide a file/device from which 32 bits will be read as a seed") },
